#include <algorithm>
#include <chrono>
#include <condition_variable>
#include <iostream>
#include <mutex>
#include <thread>
#include <vector>

using namespace std;
using namespace std::chrono;

struct ProducerConsumer {
	ProducerConsumer(int s, int c, int i) : lastLetter('a'), buffer(s, 'a'), curIdx(0), chunk(c), maxIters(i) {
		produceData = true; consumeData = false;
		thread producer(&ProducerConsumer::produce, this);
		thread consumer(&ProducerConsumer::consume, this);
		consumer.join(); producer.join();
	}

	void produce()	{
		for (unsigned long j = 0; j < maxIters; ++j)	{
			unique_lock<mutex> lck(mtx);
			prodVar.wait(lck, [this]{ return produceData; });
			produceData = false;

			char curLetter = lastLetter;
			for (unsigned long i = curIdx; i < curIdx+chunk; ++i)	{
				if (curLetter > 'z')
					curLetter = 'a';
				buffer[i % buffer.size()] = curLetter++;
				this_thread::sleep_for(milliseconds(200));
			}
			lastLetter = curLetter;

			consumeData = true;
			consVar.notify_one();
		}
	}

	void consume()	{
		for (unsigned long j = 0; j < maxIters; ++j)	{
			unique_lock<mutex> lck(mtx);
			consVar.wait(lck, [this]{ return consumeData; });
			consumeData = false;

			for (unsigned long i = curIdx; i < curIdx+chunk; ++i)	{
				cout<<buffer[i % buffer.size()]<<flush;
				this_thread::sleep_for(milliseconds(100));
			}
			curIdx = (curIdx+chunk) % buffer.size();

			produceData = true;
			prodVar.notify_one();
		}
	}

	char lastLetter;
	vector<char> buffer;
	bool produceData, consumeData;
	unsigned long curIdx, chunk, maxIters;

	mutex mtx;
	condition_variable prodVar, consVar;
};

int main()	{
	cout<<"ProducerConsumer application started..."<<endl<<endl;
	ProducerConsumer(10, 3, 40);
	cout<<endl;

	return 0;
}

